'''
Define an Aperture object, which
    -has a location
    -does photometry
    -can be plotted and picked
    -can be unplotted
'''

from ..imports import *
import photutils

class InteractiveAperture(photutils.CircularAperture):

    def __init__(self, name='', pos=(0,0), loupe=None,  aperture_radius=3):

        # initialize the CircularAperture as normal
        photutils.CircularAperture.__init__(self, pos, aperture_radius)

        # keep track of this aperture's name
        self.name = name

        self.loupe = loupe

        self.flux = np.nan

    def plot(self, ax):
        '''
        Plot this aperture onto the given axes.
        '''

        # create an empty dictionary to store any plots
        self.plotted = {}

        # (this can handle only single-position apertures)
        pos = self.positions[0]

        # (pull out the apeture radius)
        aperture_radius = self.r

        # create the circle
        self.plotted['circle'] = plt.Circle(pos, aperture_radius, color='black',fill=False,zorder=100,picker=True)
        ax.add_artist(self.plotted['circle'])
        self.plotted['marker'] = plt.scatter(pos[0], pos[1], marker='+', alpha=0.3, color='black')
        self.plotted['label'] = plt.text(pos[0]+aperture_radius+1, pos[1], self.name, va='center', zorder=11, color='black')

    def erase(self):
        '''
        Remove any traces of this aperture from the plot.
        '''

        for k in self.plotted.keys():

            self.plotted[k].remove()
            #with self.loupe.out:
            #    print('  removing {}'.format(k))

    def update(self):
        '''
        After a plot has already been created, update its hidden properties,
        as well as its appearance on the interactive plot.
        '''

        # update the radius
        self.r = self.loupe.aperture_radius

        # update the plot
        pos = self.positions[0]
        self.plotted['circle'].set_radius(self.r)
        self.plotted['label'].set_position((pos[0]+self.r+1, pos[1]))
        self.plotted['marker'].set_offsets(pos)

        # actually do the photometry
        self.phot_table = photutils.aperture_photometry(self.loupe.image, self)
        self.flux = self.phot_table['aperture_sum'].data[0]

    def __str__(self):
        '''
        How should this aperature appear as a string (for example, if it is
        included in a `'{}'.format()` statement)?
        '''
        return '{:>8} at ({:4.1f}, {:4.1f}) sums to {:10.1f} (area of {:.1f} pixels).'.format(self.name, self.positions[0][0], self.positions[0][1],  self.flux, self.area())

class ApertureWithBackground(InteractiveAperture):
    pass
    #def __init__(self, name='', pos=(0,0), aperture_radius=3 ):
    #    pass



import numpy as np, matplotlib.pyplot as plt


def create_test_array(N=5, xsize=100, ysize=100, nstars=None, single=False, seed=None):
    '''
    Create a fake stack of images of stars
    (no cosmics injected, no jitter, Gaussian PDF).

    Parameters:
    -----------
    N : int
        The number of images.
    xsize : int
        The number of columns.
    ysize : int
        The number of rows.
    nstars : int
        The number of stars to include (defaults to be \propto to image area)
    single : bool
        Should this be just a single star in the center, or should it be many?

    Returns
    -------
    image_stack : array
        (N, ysize, xsize) in shape
    '''

    # seed the random number generator
    np.random.seed(seed)

    # create images of x, y, and an empty one to fill with stars
    x1d = np.arange(0, xsize)
    y1d = np.arange(0, ysize)
    x, y = np.meshgrid(x1d, y1d)
    stars = np.zeros_like(x)

    # create N random position for stars
    if single:
        nstars = 1
        sx = np.random.normal(xsize / 2.0, 1)
        sy = np.random.normal(ysize / 2.0, 1)
    else:
        if nstars is None:
            nstars = np.minimum(int(xsize * ysize / 4), 1000)
        sx = np.random.uniform(0, xsize, nstars)
        sy = np.random.uniform(0, ysize, nstars)

    # set some background level
    bg = 30

    # create a semi-reasonable magnitude distribution for stars
    if single:
        topmag = 5
    else:
        topmag = 10
    sf = 10000 * 10**(-0.4 * np.random.triangular(0, topmag, topmag, nstars))

    # define the cartoon PSF for the stars
    sigma = 1.0

    def gauss(x, y, x0, y0):
        return np.exp(-0.5 * ((x - x0)**2 + (y - y0)**2) / sigma**2)

    # create a cube with an image for each star, and sum them together into one image
    stars = (sf * gauss(x[:, :, np.newaxis],
                        y[:, :, np.newaxis], x0=sx, y0=sy)).sum(2)

    # create a model images
    model = (bg + stars)[np.newaxis, :, :] * np.ones(N).reshape((N, 1, 1))

    # add some noise to it
    image = np.random.normal(model, np.sqrt(model))

    return image



from ..imports import *
import illumination as il
from IPython.display import display
import ipywidgets as widgets
from .apertures import *
from ..imaging import io
from astropy.table import vstack

# https://github.com/matplotlib/jupyter-matplotlib might be good for widgets!


class Loupe(il.IllustrationBase):
    '''
    An interactive Illustration, that allows us to do
    aperture photometry on particular locations in an image.
    '''
    illustrationtype = 'Loupe'

    def __init__(self, image,
                       aperture_radius=3.0,
                       **framekw):
        '''
        Parameters
        ----------

        image : 2D array, str
            An image or a path to a filename containing an image.

        aperture_radius : float
            The default radius for the photometric aperture
            (can be changed interactively).

        subtract_background : bool
            Should we attempt to do background subtraction?
            (not fully implemented yet)

        **framekw passed to CameraFrame
        '''

        # kludge
        subtract_background = False

        # load an image if given a filename
        if type(image) == str:
            image = io.read_image(image)

        # don't let this illustration talk
        self._pithy = True

        # store the image as an attribute of the loupe
        self.image = image


        # store the initial defaults
        self.aperture_radius = aperture_radius
        self.subtract_background = subtract_background

        # start with no apertures
        self.apertures = []
        self.napertures = 0

        # set up an interactive widget

        aperture_radius_slider = widgets.FloatSlider(
                                    value=self.aperture_radius,
                                    min=0,
                                    max=50,
                                    description='Radius (px)',
                                    continuous_update=False,
                                    orientiation='vertical',
                                    readout=True)

        # idea! use a FloatRangeSlider for the background annulus
        if self.subtract_background :
            self.widge = widgets.interactive(self.photometry,
                                             aperture_radius=aperture_radius_slider,
                                             r_in=(0, 30),
                                             r_out=(0, 30),
                                             back_photo=True)
        else:
            self.widge = widgets.interactive(self.photometry,
                                             aperture_radius=aperture_radius_slider,
                                             r_in=widgets.fixed(0),
                                             r_out=widgets.fixed(0),
                                             back_photo=widgets.fixed(False))

        #self.window = widgets.Output(layout={'border': '1px solid black', 'width':'70%', 'height':'500px'})
        self.out = widgets.Output(layout={'border': '1px solid black'})




        # figure out the geometry, for the figure
        aspectratio = self.image.shape[1]/float(self.image.shape[0])
        size = 5
        if aspectratio >=1:
            width = size*aspectratio
            height = size
        else:
            width = size
            height = size/float(aspectratio)

        # create the basic illustration layout (this creates a self.grid)
        il.IllustrationBase.__init__(self, 1, 1,
                                  figkw=dict(figsize=(width, height)),
                                  hspace=0.02, wspace=0.02,
                                  left=0.05, right=0.95,
                                  bottom=0.1, top=0.9)
        self._pithy = True

        # create an imshowFrame to display the image (and hide its text output)
        self.frames['imshow'] = il.imshowFrame(data=il.make_image_sequence(image), illustration=self, **framekw)
        self.frames['imshow']._pithy=True

        # create an ax for this frame
        self.frames['imshow'].ax = plt.subplot(self.grid[0, 0])



        # draw this illustration
        self.plot()

        # set it up so we can add or subtract an aperture by clicking on the image
        self.figure.canvas.mpl_connect('key_press_event', self.do_something_with_keyboard)

        print("Please [a]dd or [z]ap an aperture at the cursor location.")

        # display the widgets
        display( widgets.HBox([self.widge, self.out]))

        # (for better layout, is it possible to place the matplotlib notebook
        #  plot *inside* a widget Output? that way, I could set the layout
        # for the output to decide where the plot goes

    def do_something_with_keyboard(self, event):
        '''
        If the user clicks on the imshow, do something with that click.

        Parameters
        ----------
        event : matplotlib.backend_bases.LocationEvent
        '''
        self.lastevent = event
        if event.key.lower() == 'z':
            self.remove_aperture(event.xdata, event.ydata)
        elif event.key.lower() == 'a':
            self.add_aperture(event.xdata, event.ydata)
        self.widge.update()

    def add_aperture(self, x, y):
        '''
        Add an aperture at a particular (x, y) location.
        '''
        #with self.out:
        #    print('adding aperture at {}'.format((x,y)))
        position = np.array([x,y])
        name = 'Star {}'.format(self.napertures)

        new = InteractiveAperture(name=name,
                                  pos=position,
                                  loupe=self,
                                  aperture_radius=self.aperture_radius)
        # plot this aperture on the image
        new.plot(ax=self.frames['imshow'].ax)

        # store this new aperture in a list
        self.apertures.append(new)
        self.napertures += 1

    def remove_aperture(self, x, y):
        '''
        Remove the closest aperture to the given (x,y) location.

        Parameters
        ----------
        x, y : float
            The location where we want to remove an aperture.
        '''
        #with self.out:
        #    print('removing aperture from {}'.format((x,y)))
        positions = [a.positions[0] for a in self.apertures]
        xexisting, yexisting = np.array(positions).T
        distance = (xexisting - x)**2 + (yexisting - y)**2
        closest = np.argmin(distance)

        toremove = self.apertures.pop(closest)
        toremove.erase()


    def photometry(self, aperture_radius=10,r_in=20,r_out=30,back_photo=True):

        '''
        Create apertures around specified list of stars

        Parameters:
        -----------

        aperture_radius : float
            The radius of a circular aperture in pixel size
        r_in : float
            The inner radius of background aperture in pixel size
            (Ignore if back_photo=False)
        r_out : float
            The outer radius of background aperture in pixel size
            (Ignore if back_photo=False)

        Returns
        -------
        photometry of each star : array
        average background pixel value around each star : array (If back_photo=True)
        plots image with the aperture and centroids located for each star
        '''

        # if we have no apertures, do nothing
        if len(self.apertures) == 0:
            return None

        # pull out the plotting ax and the image
        ax = self.frames['imshow'].ax
        image = self.image

        self.aperture_radius = aperture_radius

        #if back_photo == True:
        #    back_aperture = photutils.CircularAnnulus(pos,r_in,r_out)

        for i, a in enumerate(self.apertures):
            a.update()


        #if back_photo == True:
        #    back_table = photutils.aperture_photometry(image,back_aperture)
        #    area = np.pi*(r_out**2-r_in**2)
        #    a = np.ones((np.shape(image)[0],np.shape(image)[1]))
        #    area = photutils.aperture_photometry(a,back_aperture)
        #    back_values = back_table['aperture_sum'].data/area['aperture_sum'].data
        #    return flux_values,back_values*np.pi*aperture_radius**2
        #else:
        self.out.clear_output()
        with self.out:
            self.summarize()
        self.measurements = vstack([a.phot_table for a in self.apertures])
        self.measurements['name'] = [a.name for a in self.apertures]
        self.measurements = self.measurements['name','xcenter', 'ycenter', 'aperture_sum']
        return self.measurements

    def summarize(self):
        print('With r={}px apertures:'.format(self.aperture_radius))
        for a in self.apertures:
            print(' {}'.format(a))



from lightkurve import KeplerTargetPixelFile
from .imports import *
from .tools import *

def download_kepler_tpf(star='Kepler-186',
                        quarter=1,
                        cadence='long',
                        **kw):
    '''
    This function is a wrapper to download one or more "Target Pixel Files" (TPF)
    of Kepler data. Each TPF is basically a movie; it contains a time series
    of images in a tiny little area around a particular Kepler target.

    This has been tested only on data from the main Kepler mission and K2.

    Parameters
    ----------

    star: str, int
        The name of the star whose light curve you want to download.
        This can be a name ("Kepler-42", "KOI-961") or a KIC number
        (8561063)

    quarter: str, int
        Which Kepler quarter to download?
            1, 2, 3 ... = download a specific quarter

    cadence: str
        'long' for Kepler long cadence
        'short' for Kepler short cadence

    Returns
    -------

    tpf: KeplerTargetPixelFile object
        This is a `lightkurve`-style TargetPixelFile object, which contains
        the attributes `tpf.time` (times in JD) and `tpf.flux` (the brightness
        of each pixel as a function of time), as well as lots of methods for
        analysis and plotting.
    '''

    # download a KeplerLightCurveFile from the MAST archive
    tpf = KeplerTargetPixelFile.from_archive(star,
                                             quarter=quarter,
                                             cadence=cadence,
                                             **kw)
    return tpf



from ..lightcurves import *

def test_download_kepler_lc():
    '''
    This function tests the wrapper for downloading
    a single quarter of Kepler data.
    '''

    lc = download_kepler_lc('Kepler-17')
    lc.plot(marker='.', linewidth=0)
    folded = lc.fold(1.48571092)
    folded.scatter(marker='.', linewidth=0)

def test_locate():
    '''
    This function should draw a single circle
    around every transit in the Kepler-17 light curve.
    '''

    # download a Kepler-17 light curve
    lc = download_kepler_lc('Kepler-10', quarter=1)

    # make a light curve plot
    lc.plot(marker='.', linewidth=0)

    # draw lines for the planet on top
    locations = locate_transits(lc, period=0.837491, t0=2454964.57513, color='royalblue', name='Kepler-10b')
    locations = locate_transits(lc, period=45.294301, t0=2455062.26648, color='darkorange', name='Kepler-10c')

    plt.legend(bbox_to_anchor=(1,1), loc='upper left')
    plt.tight_layout()
